#  kmeans.py
#
import sarg, random, math
from graphics import *
from sys  import maxsize

colors= ("red","green","blue","orange","yellow","purple")
ncolors = len(colors)

fig = 0

nVarieties = sarg.Int("varieties",   4)
nBuyers = sarg.Int("buyers"      ,  20)
pixels  = sarg.Int("pixels"      , 600)

win = GraphWin("Kmean's Clustering",pixels,pixels)
drawn = []

def randomXY () :
    bounds = pixels/20
    x = bounds + random.random()*(pixels-bounds)
    y = bounds + random.random()*(pixels-bounds)
    return x,y
    
class Variety :
    def __init__ (self, label) :
        self.x,self.y = randomXY()
        self.label = label
        self.buyers = []

    def attachBuyer (self, buyer) :
        self.buyers.append(buyer)

    def detachBuyer (self, buyer) :
        if buyer in self.buyers :
            self.buyers.remove(buyer)

    def centerToBuyers(self) :
        nBuyers = len(self.buyers)
        if nBuyers :
            xsum = ysum = 0.0
            for buyer in self.buyers :
                xsum += buyer.x
                ysum += buyer.y
            self.x = xsum/nBuyers
            self.y = ysum/nBuyers
    
    def distTo (self, buyer) :
        xdist = abs(self.x-buyer.x)
        ydist = abs(self.y-buyer.y)
        return math.sqrt(xdist*xdist + ydist*ydist)


class Buyer :
    def __init__ (self, label, varieties) :
        self.x,self.y = randomXY()
        self.label   = label
        self.variety = varieties[random.randint(0,len(varieties)-1)]
        self.variety.attachBuyer(self)

    def choose(self, varieties) :
        bestVariety = self.variety 
        bestDist = bestVariety.distTo(self)
        for tryout in varieties :
            if tryout != self.variety :
                tryDist = tryout.distTo(self)
                if tryDist < bestDist :
                    bestVariety = tryout
                    bestDist    = tryDist
        if bestVariety != self.variety :
            self.variety.detachBuyer(self)
            bestVariety.attachBuyer(self)
            self.variety = bestVariety
            return True                # changed variety
        return False

def getColor(label) :
    slot =  ord(label.lower())-ord('a')
    return colors[slot%len(colors)]  #nifty wrap-around

def kmeans(seed)  :
    labels = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
    blabels= labels.lower()

    varieties = [Variety(labels[i])        for i in range(nVarieties)]
    buyers = [Buyer(blabels[i], varieties) for i in range(nBuyers)]
    fig = 0
    display(varieties, buyers, "Initial Placement")
    while True :
        changed = False
        fig += 1
        for b in buyers :
            changed = b.choose(varieties) or changed
        display(varieties, buyers, "Fig %d Buyers Choose Closest"%fig)

        if not changed : break

        fig += 1
        for v in varieties :
            v.centerToBuyers()
        display(varieties, buyers, "Fig %d Varieties Move to Center"%fig)
    display(varieties, buyers, "Fig %d Done. Stability found"%fig)

def display(varieties, buyers, banner) :
    for obj in drawn : obj.undraw()

    txtBanner = Text(Point(pixels//2,20), banner)
    txtBanner.draw(win)
    drawn.append(txtBanner)

    for b in buyers :
        v = b.variety
        color = getColor(v.label)
        p1 = Point(b.x,b.y)
        p2 = Point(v.x,v.y)

        line = Line(p1,p2)
        line.setFill(color)
        line.draw(win)
        drawn.append(line)
        
        txtLabel = Text(p1, b.label)
        txtLabel.draw(win)
        drawn.append(txtLabel)

    for v in varieties :
        color = getColor(v.label)
        txtVar = Text(Point(v.x,v.y),v.label)
        txtVar.draw(win)
        drawn.append(txtVar)
    win.getMouse()

if __name__ == "__main__" :
    seed    = sarg.Int("seed", 0)
    if seed : random.seed(seed)
    else :
        seed = random.randint(101,999)
        print("Seed is %s" % seed)
        random.seed(seed)
    kmeans(seed)
